import re
from copy import deepcopy
from typing import Any, Dict, List
from urllib.parse import urljoin

import pandas as pd
import requests

from ..config import local_config
from ..exceptions import CredentialError
from ..utils import handle_api_response
from .base import Source


class CloudForCustomers(Source):
    DEFAULT_PARAMS = {"$format": "json"}

    def __init__(
        self,
        *args,
        report_url: str = None,
        url: str = None,
        endpoint: str = None,
        params: Dict[str, Any] = None,
        env: str = "QA",
        credentials: Dict[str, Any] = None,
        **kwargs,
    ):
        """Cloud for Customers connector build for fetching Odata source.
        See [pyodata docs](https://pyodata.readthedocs.io/en/latest/index.html) for an explanation
        how Odata works.

        Parameters
        ----------
            report_url (str, optional): The url to the API in case of prepared report. Defaults to None.
            url (str, optional): The url to the API. Defaults to None.
            endpoint (str, optional): The endpoint of the API. Defaults to None.
            params (Dict[str, Any]): The query parameters like filter by creation date time. Defaults to json format.
            env (str, optional): The development environments. Defaults to 'QA'.
            credentials (Dict[str, Any], optional): The credentials are populated with values from config file or this
            parameter. Defaults to None than use credentials from local_config.
        """
        super().__init__(*args, **kwargs)

        try:
            DEFAULT_CREDENTIALS = local_config["CLOUD_FOR_CUSTOMERS"].get(env)
        except KeyError:
            DEFAULT_CREDENTIALS = None
        self.credentials = credentials or DEFAULT_CREDENTIALS or {}

        self.url = url or self.credentials.get("server")
        self.report_url = report_url

        if self.url is None and report_url is None:
            raise CredentialError("One of: ('url', 'report_url') is required.")

        self.is_report = bool(report_url)
        self.query_endpoint = endpoint

        if params:
            params_merged = self.DEFAULT_PARAMS.copy()
            params_merged.update(params)

            self.params = params_merged
        else:
            self.params = self.DEFAULT_PARAMS

        if self.url:
            self.full_url = urljoin(self.url, self.query_endpoint)

        super().__init__(*args, credentials=self.credentials, **kwargs)

    @staticmethod
    def change_to_meta_url(url: str) -> str:
        start = url.split(".svc")[0]
        url_raw = url.split("?")[0]
        end = url_raw.split("/")[-1]
        meta_url = start + ".svc/$metadata?entityset=" + end
        return meta_url

    def _to_records_report(self, url: str) -> List[Dict[str, Any]]:
        """Fetches the data from source with report_url.
        At first enter url is from function parameter. At next is generated automaticaly.
        """
        records = []
        while url:
            response = self.get_response(url, params=self.params)
            response_json = response.json()
            new_records = self.response_to_entity_list(response_json, url)
            records.extend(new_records)

            url = response_json["d"].get("__next")

        return records

    def _to_records_other(self, url: str) -> List[Dict[str, Any]]:
        """Fetches the data from source with url.
        At first enter url is a join of url and endpoint passed into this function.
        At any other entering it bring `__next_url` adress, generated automatically, but without params.
        """
        tmp_full_url = deepcopy(url)
        tmp_params = deepcopy(self.params)
        records = []
        while url:
            response = self.get_response(tmp_full_url, params=tmp_params)
            response_json = response.json()
            if isinstance(response_json["d"], dict):
                # ODATA v2+ API
                new_records = response_json["d"].get("results")
                url = response_json["d"].get("__next", None)
            else:
                # ODATA v1
                new_records = response_json["d"]
                url = response_json.get("__next", None)
            # prevents concatenation of previous url's with params with the same params
            tmp_params = None
            tmp_full_url = url
            records.extend(new_records)
        return records

    def to_records(self) -> List[Dict[str, Any]]:
        """
        Download a list of entities in the records format
        """
        if self.is_report:
            url = self.report_url
            return self._to_records_report(url=url)
        else:
            url = self.full_url
            return self._to_records_other(url=url)

    def response_to_entity_list(self, dirty_json: Dict[str, Any], url: str) -> List:
        """Changing request json response to list.

        Args:
            dirty_json (Dict[str, Any]): json from response.
            url (str): the URL which trying to fetch metadata.

        Returns:
            List: List of dictionaries.
        """

        metadata_url = self.change_to_meta_url(url)
        column_maper_dict = self.map_columns(metadata_url)
        entity_list = []
        for element in dirty_json["d"]["results"]:
            new_entity = {}
            for key, object_of_interest in element.items():
                if key not in ["__metadata", "Photo", "", "Picture"]:
                    if "{" not in str(object_of_interest):
                        new_key = column_maper_dict.get(key)
                        if new_key:
                            new_entity[new_key] = object_of_interest
                        else:
                            new_entity[key] = object_of_interest
            entity_list.append(new_entity)
        return entity_list

    def map_columns(self, url: str = None) -> Dict[str, str]:
        """Fetch metadata from url used to column name map.

        Args:
            url (str, optional): the URL which trying to fetch metadata. Defaults to None.

        Returns:
            Dict[str, str]: Property Name as key mapped to the value of sap label.
        """
        column_mapping = {}
        if url:
            username = self.credentials.get("username")
            pw = self.credentials.get("password")
            response = requests.get(url, auth=(username, pw))
            for sentence in response.text.split("/>"):
                result = re.search(
                    r'(?<=Name=")([^"]+).+(sap:label=")([^"]+)+', sentence
                )
                if result:
                    key = result.groups(0)[0]
                    val = result.groups(0)[2]
                    column_mapping[key] = val
        return column_mapping

    def get_response(
        self, url: str, params: Dict[str, Any] = None, timeout: tuple = (3.05, 60 * 30)
    ) -> requests.models.Response:
        """Handle and raise Python exceptions during request. Using of url and service endpoint needs additional parameters
           stores in params. report_url contain additional params in their structure.
           In report_url scenario it can not contain params parameter.

        Args:
            url (str): the URL which trying to connect.
            params (Dict[str, Any], optional): Additional parameters like filter, used in case of normal url.
            Defaults to None used in case of report_url, which can not contain params.
            timeout (tuple, optional): the request times out. Defaults to (3.05, 60 * 30).

        Returns:
            requests.models.Response
        """
        username = self.credentials.get("username")
        pw = self.credentials.get("password")
        response = handle_api_response(
            url=url,
            params=params,
            auth=(username, pw),
            timeout=timeout,
        )
        return response

    def to_df(
        self,
        fields: List[str] = None,
        if_empty: str = "warn",
        dtype: dict = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Returns records in a pandas DataFrame.
        Args:
            fields (List[str], optional): List of fields to put in DataFrame. Defaults to None.
            dtype (dict, optional): The dtypes to use in the DataFrame. We catch this parameter here since
            pandas doesn't support passing dtypes (eg. as a dict) to the constructor.
            kwargs: The parameters to pass to DataFrame constructor.
        """
        records = self.to_records()
        df = pd.DataFrame(data=records, **kwargs)
        if dtype:
            df = df.astype(dtype)
        if fields:
            return df[fields]
        return df



import pyodbc
import pytest

from viadot.sources.azure_blob_storage import AzureBlobStorage
from viadot.sources.azure_sql import AzureSQL

SCHEMA = "sandbox"
TABLE = "test_azure_sql"
TABLE_2 = "test_azure_sql_2"


@pytest.fixture(scope="session")
def azure_sql(TEST_CSV_FILE_PATH, TEST_CSV_FILE_BLOB_PATH):
    # Upload the test file to Blob Storage.
    azstorage = AzureBlobStorage()
    azstorage.to_storage(
        from_path=TEST_CSV_FILE_PATH, to_path=TEST_CSV_FILE_BLOB_PATH, overwrite=True
    )

    azure_sql = AzureSQL(config_key="AZURE_SQL")

    yield azure_sql

    try:
        azure_sql.run(f"DROP TABLE {SCHEMA}.{TABLE}")
        azure_sql.run(f"DROP TABLE {SCHEMA}.{TABLE_2}")
    except pyodbc.ProgrammingError:
        # in case tests end prematurely
        pass


def test_connection(azure_sql):
    azure_sql.con


def test_table_exists(azure_sql):
    result = azure_sql.exists(table=TABLE, schema=SCHEMA)
    assert result == False


def test_create_table_replace(azure_sql):
    dtypes = {"country": "VARCHAR(100)", "sales": "FLOAT(24)"}
    result = azure_sql.create_table(
        schema=SCHEMA, table=TABLE, dtypes=dtypes, if_exists="replace"
    )
    assert result == True
    table_object_id = azure_sql.run(f"SELECT OBJECT_ID('{SCHEMA}.{TABLE}', 'U')")[0][0]
    assert table_object_id is not None


def test_create_table_delete(azure_sql, TEST_CSV_FILE_BLOB_PATH):
    insert_executed = azure_sql.bulk_insert(
        schema=SCHEMA,
        table=TABLE,
        source_path=TEST_CSV_FILE_BLOB_PATH,
        if_exists="replace",
    )
    assert insert_executed is True

    result = azure_sql.run(f"SELECT SUM(sales) FROM {SCHEMA}.{TABLE} AS total")
    assert int(result[0][0]) == 230

    table_object_id_insert = azure_sql.run(
        f"SELECT OBJECT_ID('{SCHEMA}.{TABLE}', 'U')"
    )[0][0]

    delete_executed = azure_sql.create_table(
        schema=SCHEMA, table=TABLE, if_exists="delete"
    )
    assert delete_executed is True

    table_object_id_delete = azure_sql.run(
        f"SELECT OBJECT_ID('{SCHEMA}.{TABLE}', 'U')"
    )[0][0]

    result = azure_sql.run(f"SELECT SUM(sales) FROM {SCHEMA}.{TABLE} AS total")
    assert result[0][0] is None
    assert table_object_id_insert == table_object_id_delete


def test_create_table_skip_1(azure_sql):
    """Test that if_exists = "skip" works when the table doesn't exist"""
    dtypes = {"country": "VARCHAR(100)", "sales": "FLOAT(24)"}
    result = azure_sql.create_table(
        schema=SCHEMA, table=TABLE_2, dtypes=dtypes, if_exists="skip"
    )
    assert result == True

    table_object_id = azure_sql.run(f"SELECT OBJECT_ID('{SCHEMA}.{TABLE}', 'U')")[0][0]
    assert table_object_id is not None


def test_create_table_skip_2(azure_sql):
    """Test that if_exists = "skip" works when the table exists"""
    dtypes = {"country": "VARCHAR(100)", "sales": "FLOAT(24)"}
    result = azure_sql.create_table(
        schema=SCHEMA, table=TABLE, dtypes=dtypes, if_exists="skip"
    )
    assert result == False


def test_bulk_insert(azure_sql, TEST_CSV_FILE_BLOB_PATH):
    azstorage = AzureBlobStorage()
    assert azstorage.exists(TEST_CSV_FILE_BLOB_PATH)

    executed = azure_sql.bulk_insert(
        schema=SCHEMA,
        table=TABLE,
        source_path=TEST_CSV_FILE_BLOB_PATH,
        if_exists="replace",
    )
    assert executed is True

    result = azure_sql.run(f"SELECT SUM(sales) FROM {SCHEMA}.{TABLE} AS total")
    assert int(result[0][0]) == 230


def test_schemas(azure_sql):
    results = azure_sql.schemas
    assert len(results) > 0


def test_tables(azure_sql):
    results = azure_sql.tables
    assert len(results) > 0



from typing import Any, Dict, List

import pandas as pd
from prefect import Task
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from viadot.sources import Outlook

logger = logging.get_logger()


class OutlookToDF(Task):
    def __init__(
        self,
        mailbox_name: str = None,
        start_date: str = None,
        end_date: str = None,
        credentials: Dict[str, Any] = None,
        output_file_extension: str = ".csv",
        limit: int = 10000,
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        self.mailbox_name = mailbox_name
        self.start_date = start_date
        self.end_date = end_date
        self.output_file_extension = output_file_extension
        self.limit = limit
        self.credentials = credentials

        super().__init__(
            name="outlook_to_csv",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download Outlook Messages to DF"""
        return super().__call__(*args, **kwargs)

    @defaults_from_attrs("mailbox_name", "start_date", "end_date", "limit")
    def run(
        self,
        mailbox_name: str,
        start_date: str = None,
        end_date: str = None,
        outbox_list: List[str] = ["Sent Items"],
        limit: int = 10000,
    ) -> pd.DataFrame:
        """
        Task for downloading data from the Outlook API to DF.

        Args:
            mailbox_name (str): Mailbox name.
            start_date (str, optional): A filtering start date parameter e.g. "2022-01-01". Defaults to None.
            end_date (str, optional): A filtering end date parameter e.g. "2022-01-02". Defaults to None.
            outbox_list (List[str], optional): List of outbox folders to differenciate between
                Inboxes and Outboxes. Defaults to ["Sent Items"].
            limit (str, optional): A limit to access last top messages. Defaults to 10_000.

        Returns:
            pd.DataFrame: The API GET as a pandas DataFrames from Outlook.
        """
        outlook = Outlook(
            credentials=self.credentials,
            mailbox_name=mailbox_name,
            start_date=start_date,
            end_date=end_date,
            limit=limit,
        )
        df = outlook.get_all_mails_to_df(outbox_list=outbox_list)
        outlook.to_csv(df)

        logger.info(
            f"Downloaded the data from the '{outlook.mailbox_name}' into the Data Frame."
        )

        return df



import json
from typing import Any, Dict, List, Literal

import pandas as pd
from prefect import Task
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities import logging

from viadot.exceptions import ValidationError
from viadot.sources import Hubspot
from viadot.task_utils import *

logger = logging.get_logger()


class HubspotToDF(Task):
    def __init__(
        self,
        hubspot_credentials: dict = None,
        hubspot_credentials_key: str = "HUBSPOT",
        *args,
        **kwargs,
    ):
        """
        Class for generating dataframe from Hubspot API endpoint.

        Args:
            hubspot_credentials (dict): Credentials to Hubspot API. Defaults to None.
            hubspot_credentials_key (str, optional): Credential key to dictionary where credentials are stored (e.g. in local config). Defaults to "HUBSPOT".

        """

        if hubspot_credentials is None:
            self.hubspot_credentials = credentials_loader.run(
                credentials_secret=hubspot_credentials_key,
            )

        else:
            self.hubspot_credentials = hubspot_credentials

        super().__init__(
            name="hubspot_to_df",
            *args,
            **kwargs,
        )

    def __call__(self):
        """Download Hubspot data to a DF"""
        super().__call__(self)

    def to_df(self, result: dict) -> pd.DataFrame:
        """
        Args:
            result (dict): API response in JSON format.

        Returns:
            pd.DataFrame: Output table generated from JSON dictionary.
        """
        return pd.json_normalize(result)

    def date_to_unixtimestamp(self, date: str = None) -> int:
        """
        Function for date conversion from user defined "yyyy-mm-dd" to Unix Timestamp (SECONDS SINCE JAN 01 1970. (UTC)).
        For example: 1680774921 SECONDS SINCE JAN 01 1970. (UTC) -> 11:55:49 AM 2023-04-06.

        Args:
            date (str, optional): Input date in format "yyyy-mm-dd". Defaults to None.

        Returns:
            int: Number of seconds that passed since 1970-01-01 until "date".
        """

        clean_date = int(datetime.timestamp(datetime.strptime(date, "%Y-%m-%d")) * 1000)
        return clean_date

    def format_filters(self, filters: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Function for API body (filters) conversion from a user defined to API language. Converts date to Unix Timestamp.

        Args:
            filters (Dict[str, Any], optional): Filters in JSON format. Defaults to {}.

        Returns:
            Dict[str, Any]: Filters in JSON format after data cleaning.
        """
        for item in filters:
            for iterator in range(len(item["filters"])):
                for subitem in item["filters"][iterator]:
                    lookup = item["filters"][iterator][subitem]
                    try:
                        date_after_format = self.date_to_unixtimestamp(lookup)
                        filters[filters.index(item)]["filters"][iterator][
                            subitem
                        ] = date_after_format
                    except:
                        pass

        return filters

    def get_offset_from_response(self, api_response: Dict[str, Any] = {}) -> tuple:
        """
        Helper funtion for assigning offset type/value depends on keys in API response.

        Args:
            api_response (Dict[str, Any], optional): API response in JSON format. Defaults to {}.

        Returns:
            tuple: Tuple in order: (offset_type, offset_value)
        """
        if "paging" in api_response.keys():
            offset_type = "after"
            offset_value = api_response["paging"]["next"][f"{offset_type}"]
        elif "offset" in api_response.keys():
            offset_type = "offset"
            offset_value = api_response["offset"]
        else:
            offset_type = None
            offset_value = None

        return (offset_type, offset_value)

    def run(
        self,
        endpoint: str,
        properties: List[Any] = [],
        filters: Dict[str, Any] = {},
        nrows: int = 1000,
    ) -> pd.DataFrame:
        """
        Run method for extraction of defined varieties of endpints for Hubspot API.

        Args:
            endpoint (str): Hubspot endpoint - full url or schema name (contacts/line_items/,...).
            properties (List[Any], optional): List of columns for the extraction. Defaults to [].
            filters (Dict[str, Any], optional): Filters in JSON format that will be passed to API body. Defaults to {}.
            nrows (int, optional): Max number of rows to pull during execution. Defaults to 1000.

        Returns:
            pd.DataFrame: Output dataframe.
        """

        hubspot = Hubspot(credentials=self.hubspot_credentials)

        url = hubspot.get_api_url(
            endpoint=endpoint,
            properties=properties,
            filters=filters,
        )

        if filters:
            filters_formatted = self.format_filters(filters=filters)
            body = hubspot.get_api_body(filters=filters_formatted)
            self.method = "POST"
            partition = hubspot.to_json(url=url, body=body, method=self.method)
            full_dataset = partition["results"]

            while "paging" in partition.keys() and len(full_dataset) < nrows:
                body = json.loads(hubspot.get_api_body(filters=filters_formatted))
                body["after"] = partition["paging"]["next"]["after"]
                partition = hubspot.to_json(
                    url=url, body=json.dumps(body), method=self.method
                )
                full_dataset.extend(partition["results"])

        else:
            self.method = "GET"
            partition = hubspot.to_json(url=url, method=self.method)
            full_dataset = partition[list(partition.keys())[0]]

            offset_type = self.get_offset_from_response(partition)[0]
            offset_value = self.get_offset_from_response(partition)[1]

            while offset_value and len(full_dataset) < nrows:
                url = hubspot.get_api_url(
                    endpoint=endpoint,
                    properties=properties,
                    filters=filters,
                )
                url += f"{offset_type}={offset_value}"

                partition = hubspot.to_json(url=url, method=self.method)
                full_dataset.extend(partition[list(partition.keys())[0]])

                offset_type = self.get_offset_from_response(partition)[0]
                offset_value = self.get_offset_from_response(partition)[1]

        df = self.to_df(full_dataset)[:nrows]

        return df



import json
from typing import Any, Dict, Literal

import pandas as pd
from prefect.utilities import logging

from ..config import local_config
from ..exceptions import APIError, CredentialError
from ..utils import handle_api_response
from .base import Source

logger = logging.get_logger(__name__)


class BusinessCore(Source):
    """
    Source for getting data from Bussines Core ERP API.

    """

    def __init__(
        self,
        url: str = None,
        filters_dict: Dict[str, Any] = {
            "BucketCount": None,
            "BucketNo": None,
            "FromDate": None,
            "ToDate": None,
        },
        verify: bool = True,
        credentials: Dict[str, Any] = None,
        config_key: str = "BusinessCore",
        *args,
        **kwargs,
    ):
        """
        Creating an instance of BusinessCore source class.

        Args:
            url (str, optional): Base url to a view in Business Core API. Defaults to None.
            filters_dict (Dict[str, Any], optional): Filters in form of dictionary. Available filters: 'BucketCount',
                'BucketNo', 'FromDate', 'ToDate'.  Defaults to {"BucketCount": None,"BucketNo": None,"FromDate": None,
                "ToDate": None,}.
            verify (bool, optional): Whether or not verify certificates while connecting to an API. Defaults to True.
            credentials (Dict[str, Any], optional): Credentials stored in a dictionary. Required credentials: username,
                password. Defaults to None.
            config_key (str, optional): Credential key to dictionary where details are stored. Defaults to "BusinessCore".

        Raises:
            CredentialError: When credentials are not found.
        """
        DEFAULT_CREDENTIALS = local_config.get(config_key)
        credentials = credentials or DEFAULT_CREDENTIALS

        required_credentials = ["username", "password"]
        if any([cred_key not in credentials for cred_key in required_credentials]):
            not_found = [c for c in required_credentials if c not in credentials]
            raise CredentialError(f"Missing credential(s): '{not_found}'.")

        self.config_key = config_key
        self.url = url
        self.filters_dict = filters_dict
        self.verify = verify

        super().__init__(*args, credentials=credentials, **kwargs)

    def generate_token(self) -> str:
        """
        Function for generating Business Core API token based on username and password.

        Returns:
        string: Business Core API token.

        """
        url = "https://api.businesscore.ae/api/user/Login"

        payload = f'grant_type=password&username={self.credentials.get("username")}&password={self.credentials.get("password")}&scope='
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        response = handle_api_response(
            url=url, headers=headers, method="GET", body=payload, verify=self.verify
        )
        token = json.loads(response.text).get("access_token")
        self.token = token
        return token

    def clean_filters_dict(self) -> Dict:
        """
        Function for replacing 'None' with '&' in a dictionary. Needed for payload in 'x-www-form-urlencoded' from.

        Returns:
        Dict: Dictionary with filters prepared for further use.
        """
        return {
            key: ("&" if val is None else val) for key, val in self.filters_dict.items()
        }

    def get_data(self) -> Dict:
        """
        Function for obtaining data in dictionary format from Business Core API.

        Returns:
        Dict: Dictionary with data downloaded from Business Core API.
        """
        filters = self.clean_filters_dict()

        payload = (
            "BucketCount="
            + str(filters.get("BucketCount"))
            + "BucketNo="
            + str(filters.get("BucketNo"))
            + "FromDate="
            + str(filters.get("FromDate"))
            + "ToDate"
            + str(filters.get("ToDate"))
        )
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": "Bearer " + self.generate_token(),
        }
        logger.info("Downloading the data...")
        response = handle_api_response(
            url=self.url,
            headers=headers,
            method="GET",
            body=payload,
            verify=self.verify,
        )
        logger.info("Data was downloaded successfully.")
        return json.loads(response.text)

    def to_df(self, if_empty: Literal["warn", "fail", "skip"] = "skip") -> pd.DataFrame:
        """
        Function for transforming data from dictionary to pd.DataFrame.

        Args:
            if_empty (Literal["warn", "fail", "skip"], optional): What to do if output DataFrame is empty. Defaults to "skip".

        Returns:
            pd.DataFrame: DataFrame with data downloaded from Business Core API view.

        Raises:
            APIError: When selected API view is not available.
        """
        view = self.url.split("/")[-1]

        if view not in [
            "GetCustomerData",
            "GetItemMaster",
            "GetPendingSalesOrderData",
            "GetSalesInvoiceData",
            "GetSalesReturnDetailData",
            "GetSalesOrderData",
            "GetSalesQuotationData",
        ]:
            raise APIError(f"View {view} currently not available.")

        data = self.get_data().get("MasterDataList")
        df = pd.DataFrame.from_dict(data)
        logger.info(
            f"Data was successfully transformed into DataFrame: {len(df.columns)} columns and {len(df)} rows."
        )
        if df.empty is True:
            self._handle_if_empty(if_empty)

        return df



from typing import Any, Dict, List, Literal

from prefect import Flow

from viadot.task_utils import df_to_parquet
from viadot.tasks import SQLServerToDF


class SQLServerToParquet(Flow):
    def __init__(
        self,
        name,
        sql_query: str,
        local_file_path: str,
        sqlserver_config_key: str = None,
        if_exists: Literal["fail", "replace", "append", "skip", "delete"] = "fail",
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for uploading data from SQL Server to Parquet file.

        Args:
            name (str, required): The name of the flow.
            sql_query (str, required): The query to execute on the SQL Server database. If don't start with "SELECT"
                returns empty DataFrame.
            local_file_path (str, required): Path to output parquet file.
            sqlserver_config_key (str, optional): The key inside local config containing the credentials. Defaults to None.
            if_exists (Literal, optional):  What to do if the file already exists. Defaults to "fail".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # SQLServerToDF
        self.sql_query = sql_query
        self.sqlserver_config_key = sqlserver_config_key
        self.timeout = timeout

        self.local_file_path = local_file_path
        self.if_exists = if_exists

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    def gen_flow(self) -> Flow:
        df_task = SQLServerToDF(timeout=self.timeout)
        df = df_task.bind(
            config_key=self.sqlserver_config_key, query=self.sql_query, flow=self
        )
        parquet = df_to_parquet.bind(
            df=df,
            path=self.local_file_path,
            if_exists=self.if_exists,
            flow=self,
        )



import datetime
from typing import Literal

import numpy as np
import pandas as pd
import prefect
from pandas import Timestamp
from prefect import Flow
from prefect.utilities import logging

from viadot.task_utils import add_ingestion_metadata_task, df_to_parquet
from viadot.tasks.azure_data_lake import AzureDataLakeUpload

logger = logging.get_logger()


class PrefectLogs(Flow):
    def __init__(
        self,
        name: str,
        query: str,
        scheduled_start_time: str = "yesterday",
        filter_type: Literal["_gte", "_gt", "_lte", "_lt"] = "_gte",
        local_file_path: str = None,
        adls_path: str = None,
        adls_sp_credentials_secret: str = None,
        vault_name: str = None,
        overwrite_adls: bool = True,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        Flow for extracting all details about flow runs that are running on Prefect.

        Args:
            name (str): The name of the flow.
            query (str): Query to be executed in Prefect API.
            scheduled_start_time (str, optional): A parameter passed to the Prefect API query. Set as 'yesterday' or date in format ex. '2022-01-01'. Defaults to 'yesterday'.
            filter_type (Literal, optional): A comparison operator passed to the Prefect API query (_gte >=, _lte <=) that refers to the left or right boundary of date ranges
                for which flow extracts data. Defaults to _gte.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_path (str, optional): Azure Data Lake destination catalog or file path. Defaults to None.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
                Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite the file in ADLS. Defaults to True.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.

            Example query:
                {
                    project {
                        id
                        name
                        flows {
                                id
                                name
                                version
                                flow_runs(
                                    order_by: {end_time: desc}
                                    where: {_and:
                                        [
                                        {scheduled_start_time:{ %s: "%s" }},
                                        {state: {_neq: "Scheduled"}}
                                        ]
                                    }
                                    )
                                        {
                                        id
                                        scheduled_start_time
                                        start_time
                                        end_time
                                        state
                                        created_by_user_id
                                        }
                        }
                    }
                }
        """

        self.name = name
        self.query = query
        self.scheduled_start_time = scheduled_start_time
        self.filter_type = filter_type
        self.local_file_path = local_file_path
        self.adls_path = adls_path
        self.vault_name = vault_name
        self.overwrite_adls = overwrite_adls
        self.timeout = timeout
        self.adls_sp_credentials_secret = adls_sp_credentials_secret

        if scheduled_start_time == "yesterday":
            self.scheduled_start_time = datetime.date.today() - datetime.timedelta(
                days=1
            )

        else:
            self.scheduled_start_time = scheduled_start_time

        super().__init__(
            name="prefect_extract_flow_logs",
            *args,
            **kwargs,
        )

        self.gen_flow()

    def get_formatted_value_from_timestamp(
        self, value_type: Literal["time", "date"], value: Timestamp
    ) -> str:
        """
        Function returns cleaned date/time from timestamp in string format (ex. "2022-01-01" or "01:23:88").
        Args:
            value_type (Literal["time", "date"], optional): "date" or "time" type extracted from Prefect Timestamp
            value (Timestamp): Timestamp value from Prefect.
        Return:
            str: date (ex. "2022-01-01") or time (ex. "01:23:88") from the Timestamp in string format
        """
        try:
            if value is not None and value_type == "date":
                return value.split("T")[0]
            elif value is not None and value_type == "time":
                return (value.split("T")[1])[:8]
        except:
            pass

    def check_if_run_authomatically(self, value: str) -> bool:
        """
        Function checks if flow has been run automatically by scheduler or triggered manually.
        Args:
            value (str): User ID created by Prefect.
            - ID = "09720c91-a99c-4f72-b7b5-3c061c83408b" refers to Prefect scheduler
            - other IDs  = tells that flow has been run manually
        Return: (bool) Boolean value True/False.
        """
        if value == "09720c91-a99c-4f72-b7b5-3c061c83408b":  # ID of a Prefect scheduler
            return True
        else:
            return False

    def gen_flow(self) -> Flow:
        query = self.query % (
            self.filter_type,
            self.scheduled_start_time,
        )

        client = prefect.Client()
        flow_runs = client.graphql(query)

        df = pd.json_normalize(
            data=flow_runs["data"]["project"],
            record_path=["flows"],
            record_prefix="flow.",
            meta=["name"],
            meta_prefix="project.",
        )

        df = df.dropna(subset=["flow.flow_runs"])
        df = df.explode("flow.flow_runs")
        df = pd.concat(
            [
                df.drop(["flow.flow_runs"], axis=1),
                df["flow.flow_runs"].apply(pd.Series),
            ],
            axis=1,
        )
        df = df.drop([0], axis=1).reset_index()

        df["start_date"] = [
            self.get_formatted_value_from_timestamp("date", df["start_time"][x])
            for x in range(df.shape[0])
        ]

        df["is_run_automatically"] = [
            self.check_if_run_authomatically(df["created_by_user_id"][x])
            for x in range(df.shape[0])
        ]

        df["scheduled_start_time_clean"] = [
            self.get_formatted_value_from_timestamp(
                "time", df["scheduled_start_time"][x]
            )
            if df["scheduled_start_time"][x] is not None
            else None
            for x in range(df.shape[0])
        ]

        df["start_time_clean"] = [
            self.get_formatted_value_from_timestamp("time", df["start_time"][x])
            if df["start_time"][x] is not None
            else None
            for x in range(df.shape[0])
        ]

        df["end_time_clean"] = [
            self.get_formatted_value_from_timestamp("time", df["end_time"][x])
            if df["end_time"][x] is not None
            else None
            for x in range(df.shape[0])
        ]

        df["time_delta"] = np.NaN

        for x in range(df.shape[0]):
            if (
                df["end_time_clean"][x] is not None
                and df["start_time_clean"][x] is not None
            ):
                df["time_delta"][x] = str(
                    datetime.timedelta(
                        seconds=(
                            datetime.datetime.strptime(
                                df["end_time_clean"][x], "%H:%M:%S"
                            )
                            - datetime.datetime.strptime(
                                df["start_time_clean"][x], "%H:%M:%S"
                            )
                        ).seconds
                    )
                )

        df["time_delta_delay"] = np.NaN

        for x in range(df.shape[0]):
            if (
                df["scheduled_start_time_clean"][x] is not None
                and df["end_time_clean"][x] is not None
            ):
                df["time_delta_delay"][x] = str(
                    datetime.timedelta(
                        seconds=(
                            datetime.datetime.strptime(
                                df["end_time_clean"][x], "%H:%M:%S"
                            )
                            - datetime.datetime.strptime(
                                df["scheduled_start_time_clean"][x], "%H:%M:%S"
                            )
                        ).seconds
                    )
                )

        df = add_ingestion_metadata_task.run(df.drop(["index"], axis=1))

        df_to_file = df_to_parquet.bind(
            df,
            path=self.local_file_path,
            flow=self,
        )

        azure_dl_upload_task = AzureDataLakeUpload(timeout=self.timeout)
        adls_upload = azure_dl_upload_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        df_to_file.set_upstream(df, flow=self)
        adls_upload.set_upstream(df_to_file, flow=self)



import os
from datetime import datetime
from typing import Any, Dict, List, Literal

import pandas as pd
from prefect import Flow, task
from prefect.engine.signals import FAIL
from prefect.triggers import all_successful
from prefect.utilities import logging

from viadot.task_utils import add_ingestion_metadata_task, adls_bulk_upload
from viadot.tasks import AzureDataLakeUpload, MindfulToCSV

logger = logging.get_logger()
file_to_adls_task = AzureDataLakeUpload()


@task
def add_timestamp(files_names: List = None, sep: str = "\t") -> None:
    """Add new column _viadot_downloaded_at_utc into each file given in the function.

    Args:
        files_names (List, optional): File names where to add the new column. Defaults to None.
        sep (str, optional): Separator type to load and to save data. Defaults to "\t".
    """
    if not files_names:
        logger.warning("Avoided adding a timestamp. No files were reported.")
    else:
        for file in files_names:
            df = pd.read_csv(file, sep=sep)
            df_updated = add_ingestion_metadata_task.run(df)
            df_updated.to_csv(file, index=False, sep=sep)


class MindfulToADLS(Flow):
    def __init__(
        self,
        name: str,
        credentials_mindful: Dict[str, Any] = None,
        credentials_secret: str = None,
        start_date: datetime = None,
        end_date: datetime = None,
        date_interval: int = 1,
        region: Literal["us1", "us2", "us3", "ca1", "eu1", "au1"] = "eu1",
        file_extension: Literal["parquet", "csv"] = "csv",
        sep: str = "\t",
        timeout: int = 3600,
        file_path: str = "",
        adls_file_path: str = None,
        adls_overwrite: bool = True,
        adls_sp_credentials_secret: str = None,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        """Mindful flow to download the CSV files and upload them to ADLS.

        Args:
            name (str): The name of the Flow.
            credentials_mindful (Dict[str, Any], optional): Credentials to connect with Mindful API. Defaults to None.
            credentials_secret (str, optional): Name of the credential secret to retreave the credentials. Defaults to None.
            start_date (datetime, optional): Start date of the request. Defaults to None.
            end_date (datetime, optional): End date of the resquest. Defaults to None.
            date_interval (int, optional): How many days are included in the request.
                If end_date is passed as an argument, date_interval will be invalidated. Defaults to 1.
            region (Literal[us1, us2, us3, ca1, eu1, au1], optional): SD region from where to interact with the mindful API. Defaults to "eu1".
            file_extension (Literal[parquet, csv], optional): File extensions for storing responses. Defaults to "csv".
            sep (str, optional): Separator in csv file. Defaults to "\t".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
            file_path (str, optional): Path where to save the file locally. Defaults to ''.
            adls_file_path (str, optional): The destination path at ADLS. Defaults to None.
            adls_overwrite (bool, optional): Whether to overwrite files in the data lake. Defaults to True.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
        """

        self.credentials_mindful = credentials_mindful
        self.credentials_secret = credentials_secret
        self.start_date = start_date
        self.end_date = end_date
        self.date_interval = date_interval
        self.region = region
        self.file_extension = file_extension
        self.sep = sep
        self.file_path = file_path
        self.timeout = timeout

        self.adls_file_path = adls_file_path
        self.adls_overwrite = adls_overwrite
        self.adls_sp_credentials_secret = adls_sp_credentials_secret

        super().__init__(*args, name=name, **kwargs)

        self.mind_flow()

    def mind_flow(self) -> Flow:
        to_csv = MindfulToCSV(timeout=self.timeout)

        file_names = to_csv.bind(
            credentials_mindful=self.credentials_mindful,
            credentials_secret=self.credentials_secret,
            start_date=self.start_date,
            end_date=self.end_date,
            date_interval=self.date_interval,
            region=self.region,
            file_extension=self.file_extension,
            file_path=self.file_path,
            flow=self,
        )

        add_timestamp.bind(file_names, sep=self.sep, flow=self)

        adls_bulk_upload.bind(
            file_names=file_names,
            file_name_relative_path=self.file_path,
            adls_file_path=self.adls_file_path,
            adls_sp_credentials_secret=self.adls_sp_credentials_secret,
            adls_overwrite=self.adls_overwrite,
            timeout=self.timeout,
            flow=self,
        )

        add_timestamp.set_upstream(file_names, flow=self)
        adls_bulk_upload.set_upstream(add_timestamp, flow=self)



import functools
import os
import re
from itertools import chain
from typing import Any, Callable, Dict, List, Literal

import pandas as pd
import prefect
import pyarrow.parquet
import pyodbc
import requests
from prefect.utilities import logging
from prefect.utilities.graphql import EnumValue, with_args
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError, HTTPError, ReadTimeout, Timeout
from requests.packages.urllib3.util.retry import Retry
from urllib3.exceptions import ProtocolError

from .exceptions import APIError
from .signals import SKIP

logger = logging.get_logger(__name__)


def slugify(name: str) -> str:
    return name.replace(" ", "_").lower()


def handle_api_response(
    url: str,
    auth: tuple = None,
    params: Dict[str, Any] = None,
    headers: Dict[str, Any] = None,
    timeout: tuple = (3.05, 60 * 30),
    method: Literal["GET", "POST", "DELETE"] = "GET",
    body: str = None,
    verify: bool = True,
) -> requests.models.Response:
    """Handle and raise Python exceptions during request with retry strategy for specific status.
    Args:
        url (str): The URL which trying to connect.
        auth (tuple, optional): Authorization information. Defaults to None.
        params (Dict[str, Any], optional): The request params also includes parameters such as the content type. Defaults to None.
        headers: (Dict[str, Any], optional): The request headers. Defaults to None.
        timeout (tuple, optional): The request times out. Defaults to (3.05, 60 * 30).
        method (Literal ["GET", "POST","DELETE"], optional): REST API method to use. Defaults to "GET".
        body (str, optional): Data to send using POST method. Defaults to None.
        verify (bool, optional): Whether to verify cerificates. Defaults to True.
    Raises:
        ValueError: Raises when 'method' parameter value hasn't been specified
        ReadTimeout: Stop waiting for a response after a given number of seconds with the timeout parameter.
        HTTPError: Exception that indicates when HTTP status codes returned values different than 200.
        ConnectionError: Exception that indicates when client is unable to connect to the server.
        ProtocolError: Raised when something unexpected happens mid-request/response.
        APIError: defined by user.
    Returns:
        requests.models.Response
    """
    if method.upper() not in ["GET", "POST", "DELETE"]:
        raise ValueError(
            f"Method not found. Please use one of the available methods: 'GET', 'POST', 'DELETE'."
        )
    try:
        session = requests.Session()
        retry_strategy = Retry(
            total=3,
            status_forcelist=[429, 500, 502, 503, 504],
            backoff_factor=1,
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        with session.request(
            url=url,
            auth=auth,
            params=params,
            headers=headers,
            timeout=timeout,
            data=body,
            method=method,
            verify=verify,
        ) as response:
            response.raise_for_status()
    except ReadTimeout as e:
        msg = "The connection was successful, "
        msg += f"however the API call to {url} timed out after {timeout[1]}s "
        msg += "while waiting for the server to return data."
        raise APIError(msg)
    except HTTPError as e:
        raise APIError(
            f"The API call to {url} failed. {e} "
            "Perhaps your account credentials need to be refreshed?",
        ) from e
    except (ConnectionError, Timeout) as e:
        raise APIError(f"The API call to {url} failed due to connection issues.") from e
    except ProtocolError as e:
        raise APIError(f"Did not receive any reponse for the API call to {url}.")
    except Exception as e:
        raise APIError("Unknown error.") from e

    return response


def get_flow_last_run_date(flow_name: str) -> str:
    """
    Retrieve a flow's last run date as an ISO datetime string.

    This function assumes you are already authenticated with Prefect Cloud.
    """
    client = prefect.Client()
    result = client.graphql(
        {
            "query": {
                with_args(
                    "flow_run",
                    {
                        "where": {
                            "flow": {"name": {"_eq": flow_name}},
                            "start_time": {"_is_null": False},
                            "state": {"_eq": "Success"},
                        },
                        "order_by": {"start_time": EnumValue("desc")},
                        "limit": 1,
                    },
                ): {"start_time"}
            }
        }
    )
    flow_run_data = result.get("data", {}).get("flow_run")

    if not flow_run_data:
        return None

    last_run_date_raw_format = flow_run_data[0]["start_time"]
    last_run_date = last_run_date_raw_format.split(".")[0] + "Z"
    return last_run_date


def get_sql_server_table_dtypes(
    table, con: pyodbc.Connection, schema: str = None
) -> dict:
    """Get column names and types from a SQL Server database table.

    Args:
        table (_type_): The table for which to fetch dtypes.
        con (pyodbc.Connection): The connection to the database where the table is located.
        schema (str, optional): The schema where the table is located. Defaults to None.

    Returns:
        dict: A dictionary of the form {column_name: dtype, column_name2: dtype2, ...}.
    """

    query = f"""
    SELECT 
        col.name,
        t.name,
        col.max_length
    FROM sys.tables AS tab
        INNER JOIN sys.columns AS col
            ON tab.object_id = col.object_id
        LEFT JOIN sys.types AS t
            ON col.user_type_id = t.user_type_id
    WHERE tab.name = '{table}'
    AND schema_name(tab.schema_id) = '{schema}'
    ORDER BY column_id;
    """
    cursor = con.cursor()
    query_result = cursor.execute(query).fetchall()
    cursor.close()

    dtypes = {}
    for row in query_result:
        column_name = row[0]
        dtype = row[1]
        length = row[2]
        if dtype == "varchar":
            dtypes[column_name] = dtype + f"({length})"
        else:
            dtypes[column_name] = dtype

    return dtypes


def _cast_df_cols(df):
    df = df.replace({"False": False, "True": True})

    datetime_cols = (col for col, dtype in df.dtypes.items() if dtype.kind == "M")
    bool_cols = (col for col, dtype in df.dtypes.items() if dtype.kind == "b")
    int_cols = (col for col, dtype in df.dtypes.items() if dtype.kind == "i")

    for col in datetime_cols:
        df[col] = df[col].dt.strftime("%Y-%m-%d %H:%M:%S+00:00")

    for col in bool_cols:
        df[col] = df[col].astype(pd.Int64Dtype())

    for col in int_cols:
        df[col] = df[col].astype(pd.Int64Dtype())

    return df


def build_merge_query(
    stg_schema: str,
    stg_table: str,
    schema: str,
    table: str,
    primary_key: str,
    con: pyodbc.Connection,
) -> str:
    """
    Build a merge query for the simplest possible upsert scenario:
    - updating and inserting all fields
    - merging on a single column, which has the same name in both tables

    Args:
        stg_schema (str): The schema where the staging table is located.
        stg_table (str): The table with new/updated data.
        schema (str): The schema where the table is located.
        table (str): The table to merge into.
        primary_key (str): The column on which to merge.
        con (pyodbc.Connection) The connection to the database on which the
        query will be executed.
    """

    # Get column names
    columns_query = f"""
    SELECT 
        col.name
    FROM sys.tables AS tab
        INNER JOIN sys.columns AS col
            ON tab.object_id = col.object_id
    WHERE tab.name = '{table}'
    AND schema_name(tab.schema_id) = '{schema}'
    ORDER BY column_id;
    """
    cursor = con.cursor()
    columns_query_result = cursor.execute(columns_query).fetchall()
    cursor.close()

    columns = [tup[0] for tup in columns_query_result]
    columns_stg_fqn = [f"stg.{col}" for col in columns]

    # Build merge query
    update_pairs = [f"existing.{col} = stg.{col}" for col in columns]
    merge_query = f"""
    MERGE INTO {schema}.{table} existing
        USING {stg_schema}.{stg_table} stg
        ON stg.{primary_key} = existing.{primary_key}
        WHEN MATCHED
            THEN UPDATE SET {", ".join(update_pairs)}
        WHEN NOT MATCHED
            THEN INSERT({", ".join(columns)})
            VALUES({", ".join(columns_stg_fqn)});
    """
    return merge_query


def gen_bulk_insert_query_from_df(
    df: pd.DataFrame, table_fqn: str, chunksize=1000, **kwargs
) -> str:
    """
    Converts a DataFrame to a bulk INSERT query.

    Args:
        df (pd.DataFrame): The DataFrame which data should be put into the INSERT query.
        table_fqn (str): The fully qualified name (schema.table) of the table to be inserted into.

    Returns:
        str: A bulk insert query that will insert all data from `df` into `table_fqn`.

    Examples:
    >>> data = [(1, "_suffixnan", 1), (2, "Noneprefix", 0), (3, "fooNULLbar", 1, 2.34)]
    >>> df = pd.DataFrame(data, columns=["id", "name", "is_deleted", "balance"])
    >>> df
       id        name  is_deleted  balance
    0   1  _suffixnan           1      NaN
    1   2  Noneprefix           0      NaN
    2   3  fooNULLbar           1     2.34
    >>> query = gen_bulk_insert_query_from_df(df, "users", status="APPROVED", address=None)
    >>> print(query)
    INSERT INTO users (id, name, is_deleted, balance, status, address)
    VALUES (1, '_suffixnan', 1, NULL, 'APPROVED', NULL),
           (2, 'Noneprefix', 0, NULL, 'APPROVED', NULL),
           (3, 'fooNULLbar', 1, 2.34, 'APPROVED', NULL);
    """
    if df.shape[1] == 1:
        raise NotImplementedError(
            "Currently, this function only handles DataFrames with at least two columns."
        )

    def _gen_insert_query_from_records(records: List[tuple]) -> str:
        tuples = map(str, tuple(records))

        # Change Nones to NULLs
        none_nan_pattern = r"(?<=\W)(nan|None)(?=\W)"
        values = re.sub(none_nan_pattern, "NULL", (",\n" + " " * 7).join(tuples))

        # Change the double quotes into single quotes, as explained above.
        # Note this pattern should be improved at a later time to cover more edge cases.
        double_quotes_pattern = r'(")(.*)(")(\)|,)'
        values_clean = re.sub(double_quotes_pattern, r"'\2'\4", values)
        # Hacky - replaces starting and ending double quotes.
        values_clean = (
            values.replace('",', "',")
            .replace(', "', ", '")
            .replace('("', "('")
            .replace("\\'", "")
        )

        return f"INSERT INTO {table_fqn} ({columns})\n\nVALUES {values_clean}"

    df = df.copy().assign(**kwargs)
    df = _cast_df_cols(df)

    columns = ", ".join(df.columns)

    tuples_raw = df.itertuples(index=False, name=None)
    # Escape values with single quotes inside by adding another single quote
    # ("val'ue" -> "val''ue").
    # As Python wraps such strings in double quotes, which are interpreted as
    # column names by SQL Server, we later also replace the double quotes with
    # single quotes.
    tuples_escaped = [
        tuple(
            f"""{value.replace("'", "''")}""" if type(value) == str else value
            for value in row
        )
        for row in tuples_raw
    ]

    if len(tuples_escaped) > chunksize:
        insert_query = ""
        chunk_start = 0
        for chunk_end in range(chunksize, len(tuples_escaped), chunksize):
            chunk = tuples_escaped[chunk_start:chunk_end]
            chunk_start += chunksize
            if len(tuples_escaped) - chunk_end < chunksize:
                chunk = tuples_escaped[chunk_end:]
            chunk_insert_query = _gen_insert_query_from_records(chunk)
            insert_query += chunk_insert_query + ";\n\n"
        return insert_query
    else:
        return _gen_insert_query_from_records(tuples_escaped)


def union_dict(*dicts):
    """
    Function that union list of dictionaries

    Args:
        dicts (List[Dict]): list of dictionaries with credentials.

    Returns:
        Dict: A single dictionary createb by union method.

    Examples:

    >>> a = {"a":1}
    >>> b = {"b":2}
    >>> union_credentials_dict(a ,b)
    {'a': 1, 'b': 2}

    """
    return dict(chain.from_iterable(dct.items() for dct in dicts))


def handle_if_empty_file(
    if_empty: Literal["warn", "skip", "fail"] = "warn",
    message: str = None,
):
    """
    Task for handling empty file.
    Args:
        if_empty (Literal, optional): What to do if file is empty. Defaults to "warn".
        message (str, optional): Massage to show in warning and error messages. Defaults to None.
    Raises:
        ValueError: If `if_empty` is set to `fail`.
        SKIP: If `if_empty` is set to `skip`.
    """
    if if_empty == "warn":
        logger.warning(message)
    elif if_empty == "skip":
        raise SKIP(message)
    elif if_empty == "fail":
        raise ValueError(message)


def check_if_empty_file(
    path: str,
    if_empty: Literal["warn", "skip", "fail"] = "warn",
):
    """
    Task for checking if the file is empty and handling it.

    Args:
        path (str, required): Path to the local file.
        if_empty (Literal, optional): What to do if file is empty. Defaults to "warn".

    """
    if_empty = if_empty or "warn"
    if os.stat(path).st_size == 0:
        handle_if_empty_file(if_empty, message=f"Input file - '{path}' is empty.")

    # when Parquet file contains only metadata
    elif os.path.splitext(path)[1] == ".parquet":
        if pyarrow.parquet.read_metadata(path).num_columns == 0:
            handle_if_empty_file(if_empty, message=f"Input file - '{path}' is empty.")


def add_viadot_metadata_columns(source_name: str = None) -> Callable:
    """
    Decorator that adds metadata columns to df in 'to_df' method.
    For now only _viadot_source is available because _viadot_downloaded_at_utc is added on the Flow level.

    Args:
        source_name (str, optional): The name of the source to be included in the DataFrame.
        This should be provided when creating a DataFrame in a Task, rather than in a Source.
        Defaults to None.

    Warning: Please remember to include brackets when applying a decorator, even if you are not passing the 'source_name' parameter.

    Example:

        In task:

        @add_viadot_metadata_columns(source_name="Sharepoint")
        def to_df(self):
            ...

        In source:

        @add_viadot_metadata_columns()
        def to_df(self):
            ...
    """

    def decorator(func) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> pd.DataFrame:
            df = func(*args, **kwargs)

            df["_viadot_source"] = (
                source_name if source_name is not None else args[0].__class__.__name__
            )

            return df

        return wrapper

    return decorator



import logging
import os
import shutil
from pathlib import Path
from typing import Any, Dict, Tuple

import great_expectations as ge
import pandas as pd
from great_expectations.data_context import BaseDataContext
from great_expectations.data_context.types.base import (
    DataContextConfig,
    DatasourceConfig,
    FilesystemStoreBackendDefaults,
)
from great_expectations.validation_operators.types.validation_operator_result import (
    ValidationOperatorResult,
)
from prefect.engine import signals
from prefect.tasks.great_expectations import RunGreatExpectationsValidation
from prefect.utilities.tasks import defaults_from_attrs


# Simplifies the setup and adds logging to the standard Prefect task
class RunGreatExpectationsValidation(RunGreatExpectationsValidation):
    """
    Task for running data validation with Great Expectations on a pandas DataFrame.
    See https://docs.prefect.io/api/latest/tasks/great_expectations.html#rungreatexpectationsvalidation
    for full documentation.

    Args:
        expectations_path (str): The path to the directory containing the expectation suites.
        df (pd.DataFrame): The DataFrame to validate.
    """

    def __init__(
        self,
        df: pd.DataFrame = None,
        expectations_path: str = None,
        evaluation_parameters: Dict[str, Any] = None,
        keep_output: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.df = df
        self.expectations_path = expectations_path
        self.evaluation_parameters = evaluation_parameters
        self.keep_output = keep_output

    @staticmethod
    def _get_batch_kwargs(df: pd.DataFrame) -> dict:
        dataset = ge.from_pandas(df)
        return {"dataset": dataset, "datasource": "pandas"}

    @staticmethod
    def _get_ge_context_local(ge_project_path: str) -> BaseDataContext:
        """
        This is configured to work with an in-memory pandas DataFrame.
        This setup allows us to run validations before (perhaps unnecessarily) writing any data
        to disk, as well as at any other stage.

        Currently using local storage.

        Args:
        ge_project_path (str): The path to the Great Expectations project,
        eg. `/home/viadot/my_flow`. Expectation suites need to be placed inside the
        `expectations` folder, eg. `/home/viadot/my_flow/expectations/failure.json`.

        Returns:
            BaseDataContext: The GE context (ie. config) required to run the validations.
        """
        data_context_config = DataContextConfig(
            datasources={
                "pandas": DatasourceConfig(
                    class_name="PandasDatasource",
                    batch_kwargs_generators={},  # override the CSV default
                )
            },
            store_backend_defaults=FilesystemStoreBackendDefaults(ge_project_path),
            validation_operators={
                "action_list_operator": {
                    "class_name": "ActionListValidationOperator",
                    "action_list": [
                        {
                            "name": "store_validation_result",
                            "action": {"class_name": "StoreValidationResultAction"},
                        },
                        {
                            "name": "store_evaluation_params",
                            "action": {"class_name": "StoreEvaluationParametersAction"},
                        },
                        {
                            "name": "update_data_docs",
                            "action": {"class_name": "UpdateDataDocsAction"},
                        },
                    ],
                }
            },
        )
        context = BaseDataContext(project_config=data_context_config)
        return context

    @defaults_from_attrs(
        "df", "expectations_path", "evaluation_parameters", "keep_output"
    )
    def run(
        self,
        df: pd.DataFrame = None,
        expectations_path: str = None,
        keep_output: bool = None,
        **kwargs,
    ):
        ge_project_path = str(Path(expectations_path).parent)

        batch_kwargs = self._get_batch_kwargs(df)
        context = self._get_ge_context_local(ge_project_path)

        self.logger.info("Beginning validation run...")

        try:
            results = super().run(
                batch_kwargs=batch_kwargs,  # input data
                context=context,  # ~project config
                **kwargs,
            )
        except signals.FAIL as e:
            results = e.state.result

        # Show summary of results
        n_successful, n_expectations = self._get_stats_from_results(results)
        status = "success" if results.success else "failure"
        level = logging.INFO if results.success else logging.ERROR
        self.logger.log(
            msg=f"Validation finished with status '{status}'. {n_successful}/{n_expectations} test(s) passed.",
            level=level,
        )

        validation_ids = [res for res in results["run_results"]]
        validation_id = validation_ids[0]
        url_dicts = context.get_docs_sites_urls(resource_identifier=validation_id)
        validation_site_url = url_dicts[0]["site_url"]

        if keep_output:
            docs_msg = f"To explore the docs, visit {validation_site_url}"
            docs_msg += " or the 'Artifacts' tab on the Prefect flow run dashboard."
            self.logger.info(docs_msg)

        else:
            docs_path = os.path.join(ge_project_path, "uncommitted")
            checkpoints_path = os.path.join(ge_project_path, "checkpoints")

            shutil.rmtree(docs_path)
            shutil.rmtree(checkpoints_path)

        if not results.success:
            raise signals.FAIL(result=results)

        return results

    def _get_stats_from_results(
        self, result: ValidationOperatorResult
    ) -> Tuple[int, int]:
        """Returns Tuple containing number of successful and evaluated expectations"""
        result_identifier = result.list_validation_result_identifiers()[0]
        stats = result._list_validation_statistics()[result_identifier]
        n_successful = stats["successful_expectations"]
        n_expectations = stats["evaluated_expectations"]
        return n_successful, n_expectations



import pandas as pd
from prefect import Task

from ..sources import Eurostat


class EurostatToDF(Task):
    """Task for creating pandas data frame from Eurostat HTTPS REST API (no credentials required).

    Args:
        dataset_code (str): The code of eurostat dataset that we would like to upload.
        params (Dict[str], optional):
            A dictionary with optional URL parameters. The key represents the parameter id, while the value is the code
            for a specific parameter, for example 'params = {'unit': 'EUR'}' where "unit" is the parameter that you would like to set
            and "EUR" is the code of the specific parameter. You can add more than one parameter, but only one code per parameter!
            So you CAN NOT provide list of codes as in example 'params = {'unit': ['EUR', 'USD', 'PLN']}'
            This parameter is REQUIRED in most cases to pull a specific dataset from the API.
            Both parameter and code has to provided as a string! Defaults to None.
        base_url (str): The base URL used to access the Eurostat API. This parameter specifies the root URL for all requests made to the API.
            It should not be modified unless the API changes its URL scheme.
            Defaults to "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/"
        requested_columns (List[str], optional): list of needed names of columns. Names should be given as str's into the list.
            Defaults to None.
    Raises:
        TypeError: If self.requested_columns have different type than a list.
    """

    def __init__(
        self,
        dataset_code: str,
        params: dict = None,
        base_url: str = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/",
        requested_columns: list = None,
        *args,
        **kwargs,
    ):
        self.dataset_code = dataset_code
        self.params = params
        self.base_url = base_url
        self.requested_columns = requested_columns
        if (
            not isinstance(self.requested_columns, list)
            and self.requested_columns is not None
        ):
            raise TypeError("Requested columns should be provided as list of strings.")

        super().__init__(name="eurostat_to_df", *args, **kwargs)

    def run(self) -> pd.DataFrame:
        """Run function for returning unchanged DataFrame, or modify DataFrame and returning if user need specific columns.

        Raises:
            ValueError: If self.requested_columns contains columns names that do not exist in the DataFrame.

        Returns:
            pd.DataFrame: Unchanged DataFrame or DataFrame with only choosen columns.
        """

        data_frame = Eurostat(
            self.dataset_code, self.params
        ).get_data_frame_from_response()

        if self.requested_columns is None:
            return data_frame
        else:
            columns_list = data_frame.columns.tolist()
            columns_list = [str(column).casefold() for column in columns_list]
            needed_column_after_validation = []
            non_available_columns = []

            for column in self.requested_columns:
                # Checking if user column is in our dataframe column list
                column = str(column).casefold()

                if column in columns_list:
                    needed_column_after_validation.append(column)
                else:
                    non_available_columns.append(column)

            # Error logger
            if non_available_columns:
                self.logger.error(
                    f"Name of the columns: '{' | '.join(non_available_columns)}' are not in DataFrame. Please check spelling!\n"
                    f"Available columns: {' | '.join(columns_list)}"
                )
                raise ValueError("Provided columns are not available!")

            new_df = data_frame.loc[:, needed_column_after_validation]
            return new_df



import struct
from datetime import datetime, timedelta, timezone
from typing import List

from .base import SQL


class SQLServer(SQL):
    DEFAULT_SCHEMA = "dbo"

    def __init__(
        self,
        config_key="SQL_SERVER",
        driver="ODBC Driver 17 for SQL Server",
        *args,
        **kwargs,
    ):
        super().__init__(*args, driver=driver, config_key=config_key, **kwargs)
        self.con.add_output_converter(-155, self._handle_datetimeoffset)

    @property
    def schemas(self) -> List[str]:
        """Returns list of schemas"""
        schemas_tuples = self.run("SELECT s.name as schema_name from sys.schemas s")
        return [schema_tuple[0] for schema_tuple in schemas_tuples]

    @property
    def tables(self) -> List[str]:
        """Returns list of tables"""
        tables_tuples = self.run(
            "SELECT schema_name(t.schema_id), t.name FROM sys.tables t"
        )
        return [".".join(row) for row in tables_tuples]

    @staticmethod
    def _handle_datetimeoffset(dto_value):
        """
        Adds support for SQL Server's custom `datetimeoffset` type, which is not
        handled natively by ODBC/pyodbc.

        See: https://github.com/mkleehammer/pyodbc/issues/134#issuecomment-281739794
        """
        (
            year,
            month,
            day,
            hour,
            minute,
            second,
            nanoseconds,
            offset_hours,
            offset_minutes,
        ) = struct.unpack("<6hI2h", dto_value)
        dt = datetime(
            year,
            month,
            day,
            hour,
            minute,
            second,
            nanoseconds // 1000,
            tzinfo=timezone(timedelta(hours=offset_hours, minutes=offset_minutes)),
        )
        return dt

    def exists(self, table: str, schema: str = None) -> bool:
        """Check whether a table exists.
        Args:
            table (str): The table to be checked.
            schema (str, optional): The schema whethe the table is located. Defaults to 'dbo'.
        Returns:
            bool: Whether the table exists.
        """

        if not schema:
            schema = self.DEFAULT_SCHEMA

        list_table_info_query = f"""
            SELECT *
            FROM sys.tables t
            JOIN sys.schemas s
                ON t.schema_id = s.schema_id
            WHERE s.name = '{schema}' AND t.name = '{table}'
        """
        exists = bool(self.run(list_table_info_query))
        return exists



import os

import pandas as pd
import pytest

from viadot.tasks import SQLiteInsert, SQLiteSQLtoDF
from viadot.tasks.sqlite import SQLiteInsert, SQLiteQuery, SQLiteSQLtoDF

TABLE = "test_sqlite_task"
DB_PATH = "testfile.sqlite"
SQL_PATH_SELECT = "testfile_select.sql"
SQL_PATH_NOTSELECT = "testfile_notselect.sql"

sqlite_insert_task = SQLiteInsert(db_path=DB_PATH)
sql_to_df_task = SQLiteSQLtoDF(db_path=DB_PATH)
sqlite_query_task = SQLiteQuery(db_path=DB_PATH)


@pytest.fixture(scope="session")
def create_test_sql_file_select():
    with open(SQL_PATH_SELECT, "w") as sql_file:
        sql_file.write(f"SELECT * FROM {TABLE};")
    yield
    os.remove(SQL_PATH_SELECT)


@pytest.fixture(scope="session")
def create_test_sql_file_notselect():
    with open(SQL_PATH_NOTSELECT, "w") as sql_file:
        sql_file.write("INSERT ...")
    yield
    os.remove(SQL_PATH_NOTSELECT)


def test_query_select(create_test_sql_file_select):
    dtypes = {"country": "VARCHAR(100)", "sales": "FLOAT(24)"}
    df_data = pd.DataFrame({"country": ["italy"], "sales": [100.0]})
    result = sqlite_insert_task.run(
        table_name=TABLE,
        dtypes=dtypes,
        df=df_data,
        if_exists="replace",
    )
    table_from_query = sql_to_df_task.run(sql_path=SQL_PATH_SELECT)
    assert result is True
    assert not table_from_query.empty


def test_query_not_select(create_test_sql_file_notselect):
    table_from_query = sql_to_df_task.run(sql_path=SQL_PATH_NOTSELECT)
    assert table_from_query.empty



from datetime import timedelta
from typing import Any, Dict, List, Union

import pandas as pd
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from ..sources import Supermetrics


class SupermetricsToCSV(Task):
    """
    Task to downloading data from Supermetrics API to CSV file.

    Args:
        path (str, optional): The destination path. Defaults to "supermetrics_extract.csv".
        max_retries (int, optional): The maximum number of retries. Defaults to 5.
        retry_delay (timedelta, optional): The delay between task retries. Defaults to 10 seconds.
        timeout (int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
        max_rows (int, optional): Maximum number of rows the query results should contain. Defaults to 1 000 000.
        max_cols (int, optional): Maximum number of columns the query results should contain. Defaults to None.
        if_exists (str, optional): What to do if file already exists. Defaults to "replace".
        if_empty (str, optional): What to do if query returns no data. Defaults to "warn".
        sep (str, optional): The separator in a target csv file. Defaults to "/t".

    """

    def __init__(
        self,
        *args,
        path: str = "supermetrics_extract.csv",
        max_retries: int = 5,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        max_rows: int = 1_000_000,
        if_exists: str = "replace",
        if_empty: str = "warn",
        sep="\t",
        **kwargs,
    ):
        self.path = path
        self.max_rows = max_rows
        self.if_exists = if_exists
        self.if_empty = if_empty
        self.sep = sep

        super().__init__(
            name="supermetrics_to_csv",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self):
        """Download Supermetrics data to a CSV"""
        super().__call__(self)

    @defaults_from_attrs(
        "path",
        "max_rows",
        "if_exists",
        "if_empty",
        "max_retries",
        "retry_delay",
        "timeout",
        "sep",
    )
    def run(
        self,
        path: str = None,
        ds_id: str = None,
        ds_accounts: Union[str, List[str]] = None,
        ds_segments: List[str] = None,
        ds_user: str = None,
        fields: List[str] = None,
        date_range_type: str = None,
        start_date: str = None,
        end_date: str = None,
        settings: Dict[str, Any] = None,
        filter: str = None,
        max_rows: int = None,
        max_columns: int = None,
        order_columns: str = None,
        if_exists: str = None,
        if_empty: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
        timeout: int = None,
        sep: str = None,
    ):
        """
        Task run method.

        Args:
            path (str, optional): The destination path. Defaulrs to None
            ds_id (str, optional): A Supermetrics query parameter.
            ds_accounts (Union[str, List[str]], optional): A Supermetrics query parameter. Defaults to None.
            ds_segments (List[str], optional): A Supermetrics query parameter. Defaults to None.
            ds_user (str, optional): A Supermetrics query parameter. Defaults to None.
            fields (List[str], optional): A Supermetrics query parameter. Defaults to None.
            date_range_type (str, optional): A Supermetrics query parameter. Defaults to None.
            start_date (str, optional): A Supermetrics query parameter. Defaults to None.
            end_date (str, optional) A Supermetrics query parameter. Defaults to None.
            settings (Dict[str, Any], optional): A Supermetrics query parameter. Defaults to None.
            filter (str, optional): A Supermetrics query parameter. Defaults to None.
            max_rows (int, optional): A Supermetrics query parameter. Defaults to None.
            max_columns (int, optional): A Supermetrics query parameter. Defaults to None.
            order_columns (str, optional): A Supermetrics query parameter. Defaults to None.
            if_exists (str, optional): What to do if file already exists. Defaults to "replace".
            if_empty (str, optional): What to do if query returns no data. Defaults to "warn".
            max_retries (int, optional): The maximum number of retries. Defaults to 5.
            retry_delay (timedelta, optional): The delay between task retries. Defaults to 10 seconds.
            timeout (int, optional): Task timeout. Defaults to 30 minuntes.
            sep (str, optional)

        """

        if max_retries:
            self.max_retries = max_retries

        if retry_delay:
            self.retry_delay = retry_delay

        if isinstance(ds_accounts, str):
            ds_accounts = [ds_accounts]

        # Build the URL
        # Note the task accepts only one account per query
        query = dict(
            ds_id=ds_id,
            ds_accounts=ds_accounts,
            ds_segments=ds_segments,
            ds_user=ds_user,
            fields=fields,
            date_range_type=date_range_type,
            start_date=start_date,
            end_date=end_date,
            settings=settings,
            filter=filter,
            max_rows=max_rows,
            max_columns=max_columns,
            order_columns=order_columns,
        )
        query = {param: val for param, val in query.items() if val is not None}
        supermetrics = Supermetrics()
        supermetrics.query(query)

        # Download data to a local CSV file
        self.logger.info(f"Downloading data to {path}...")
        supermetrics.to_csv(path, if_exists=if_exists, if_empty=if_empty, sep=sep)
        self.logger.info(f"Successfully downloaded data to {path}.")


class SupermetricsToDF(Task):
    """
    Task for downloading data from the Supermetrics API to a pandas DataFrame.

    Args:
        ds_id (str, optional): A Supermetrics query parameter.
        ds_accounts (Union[str, List[str]], optional): A Supermetrics query parameter. Defaults to None.
        ds_segments (List[str], optional): A Supermetrics query parameter. Defaults to None.
        ds_user (str, optional): A Supermetrics query parameter. Defaults to None.
        fields (List[str], optional): A Supermetrics query parameter. Defaults to None.
        date_range_type (str, optional): A Supermetrics query parameter. Defaults to None.
        settings (Dict[str, Any], optional): A Supermetrics query parameter. Defaults to None.
        filter (str, optional): A Supermetrics query parameter. Defaults to None.
        max_rows (int, optional): A Supermetrics query parameter. Defaults to None.
        max_columns (int, optional): A Supermetrics query parameter. Defaults to None.
        order_columns (str, optional): A Supermetrics query parameter. Defaults to None.
        if_empty (str, optional): What to do if query returns no data. Defaults to "warn".
        max_retries (int, optional): The maximum number of retries. Defaults to 5.
        retry_delay (timedelta, optional): The delay between task retries. Defaults to 10 seconds.
        timeout (int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        *args,
        if_empty: str = "warn",
        max_rows: int = 1_000_000,
        max_retries: int = 5,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        **kwargs,
    ):
        self.if_empty = if_empty
        self.max_rows = max_rows

        super().__init__(
            name="supermetrics_to_df",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs(
        "if_empty",
        "max_rows",
        "max_retries",
        "retry_delay",
        "timeout",
    )
    def run(
        self,
        ds_id: str = None,
        ds_accounts: Union[str, List[str]] = None,
        ds_segments: List[str] = None,
        ds_user: str = None,
        fields: List[str] = None,
        date_range_type: str = None,
        start_date: str = None,
        end_date: str = None,
        settings: Dict[str, Any] = None,
        filter: str = None,
        max_rows: int = None,
        max_columns: int = None,
        order_columns: str = None,
        if_empty: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
        timeout: int = None,
    ) -> pd.DataFrame:
        """
        Task run method.

        Args:
            ds_id (str, optional): A Supermetrics query parameter.
            ds_accounts (Union[str, List[str]], optional): A Supermetrics query parameter. Defaults to None.
            ds_segments (List[str], optional): A Supermetrics query parameter. Defaults to None.
            ds_user (str, optional): A Supermetrics query parameter. Defaults to None.
            fields (List[str], optional): A Supermetrics query parameter. Defaults to None.
            date_range_type (str, optional): A Supermetrics query parameter. Defaults to None.
            start_date (str, optional): A query paramter to pass start date to the date range filter. Defaults to None.
            end_date (str, optional): A query paramter to pass end date to the date range filter. Defaults to None.
            settings (Dict[str, Any], optional): A Supermetrics query parameter. Defaults to None.
            filter (str, optional): A Supermetrics query parameter. Defaults to None.
            max_rows (int, optional): A Supermetrics query parameter. Defaults to None.
            max_columns (int, optional): A Supermetrics query parameter. Defaults to None.
            order_columns (str, optional): A Supermetrics query parameter. Defaults to None.
            if_empty (str, optional): What to do if query returns no data. Defaults to "warn".
            max_retries (int, optional): The maximum number of retries. Defaults to 5.
            retry_delay (timedelta, optional): The delay between task retries. Defaults to 10 seconds.
            timeout (int, optional): Task timeout. Defaults to 30 minuntes.

        Returns:
            pd.DataFrame: The query result as a pandas DataFrame.
        """

        if max_retries:
            self.max_retries = max_retries

        if retry_delay:
            self.retry_delay = retry_delay

        if isinstance(ds_accounts, str):
            ds_accounts = [ds_accounts]

        # Build the URL
        # Note the task accepts only one account per query
        query = dict(
            ds_id=ds_id,
            ds_accounts=ds_accounts,
            ds_segments=ds_segments,
            ds_user=ds_user,
            fields=fields,
            date_range_type=date_range_type,
            start_date=start_date,
            end_date=end_date,
            settings=settings,
            filter=filter,
            max_rows=max_rows,
            max_columns=max_columns,
            order_columns=order_columns,
        )
        query = {param: val for param, val in query.items() if val is not None}
        supermetrics = Supermetrics()
        supermetrics.query(query)

        # Download data to a local CSV file
        self.logger.info(f"Downloading data to a DataFrame...")
        df = supermetrics.to_df(if_empty=if_empty)
        self.logger.info(f"Successfully downloaded data to a DataFrame.")
        return df
